/*
 * SNES9xGraphics.cpp
 *
 *  Created on: Nov 14, 2010
 *      Author: Squarepusher2
 */

#include "snes9x.h"
#include "SNES9xGraphics.h"

#include <assert.h>

#include "cellframework/logger/Logger.h"

SNES9xGraphics::SNES9xGraphics() : PSGLGraphics(), gl_buffer(NULL), vertex_buf(NULL)
{
   	LOG("SNES9xGraphics constructor\n");
	overscan = false;
	overscan_amount = 0.0;
}

SNES9xGraphics::~SNES9xGraphics()
{
   	LOG("SNES9xGraphics destructor\n");
	Deinit();
}

void SNES9xGraphics::Deinit()
{
   	LOG("SNES9xGraphics::Deinit()\n");
	PSGLGraphics::Deinit();
	if(vertex_buf)
	{
		free(vertex_buf);
		vertex_buf = NULL;
	}
	if(gl_buffer)
	{
		free(gl_buffer);
		gl_buffer = NULL;
	}
}

void SNES9xGraphics::DeInitDbgFont()
{
	PSGLGraphics::DeinitDbgFont();
}

void SNES9xGraphics::InitDbgFont()
{
	PSGLGraphics::InitDbgFont();
}

void SNES9xGraphics::RenderDbgFont()
{
   	LOG("SNES9xGraphics::RenderDbgFont()\n");
	cellDbgFontDraw();
}

void SNES9xGraphics::Swap()
{
   	LOG("SNES9xGraphics::Swap()\n");
	psglSwap();
}


CellVideoOutState SNES9xGraphics::GetVideoOutState()
{
   	LOG("SNES9xGraphics::GetVideoState()\n");
	return stored_video_state;
}

void SNES9xGraphics::Clear()
{
   	LOG("SNES9xGraphics::Clear()\n");
	glClear(GL_COLOR_BUFFER_BIT);
}

void SNES9xGraphics::SetViewports()
{
   LOG("SNES9xGraphics::SetViewPorts()\n");
   float device_aspect = this->GetDeviceAspectRatio();
   GLuint width = this->GetResolutionWidth();
   GLuint height = this->GetResolutionHeight();
   
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   // calculate the glOrtho matrix needed to transform the texture to the desired aspect ratio
   float desired_aspect = Settings.PS3KeepAspect ? SCREEN_REAL_ASPECT_RATIO : SCREEN_16_9_ASPECT_RATIO;

   GLuint real_width = width, real_height = height;

   // If the aspect ratios of screen and desired aspect ratio are sufficiently equal (floating point stuff), 
   // assume they are actually equal.
   if ( (int)(device_aspect*1000) > (int)(desired_aspect*1000) )
   {
      float delta = (desired_aspect / device_aspect - 1.0) / 2.0 + 0.5;
      //glOrthof(0.5 - delta, 0.5 + delta, 0, 1, -1, 1);
      glViewport(width * (0.5 - delta), 0, 2.0 * width * delta, height);
      real_width = (int)(2.0 * width * delta);
   }

   else if ( (int)(device_aspect*1000) < (int)(desired_aspect*1000) )
   {
      float delta = (device_aspect / desired_aspect - 1.0) / 2.0 + 0.5;
      //glOrthof(0, 1, 0.5 - delta, 0.5 + delta, -1, 1);
      glViewport(0, height * (0.5 - delta), width, 2.0 * height * delta);
      real_height = (int)(2.0 * height * delta);
   }
   else
      glViewport(0, 0, width, height);

	//glViewport(0, 0, width, height);
   if (overscan)
   {
      glOrthof(-overscan_amount/2, 1 + overscan_amount/2, -overscan_amount/2, 1 + overscan_amount/2, -1, 1);
   }
   else
   {
      glOrthof(0, 1, 0, 1, -1, 1);
   }

   cg_viewport_width = real_width;
   cg_viewport_height = real_height;
   
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   cgGLSetStateMatrixParameter(ModelViewProj_cgParam, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
}

void SNES9xGraphics::SetOverscan(bool will_overscan, float amount)
{
   	LOG("SNES9xGraphics::SetOverscan(bool will_overscan, float amount)()\n");
	overscan_amount = amount;
	overscan = will_overscan;
	SetViewports();
}

void SNES9xGraphics::SetAspectRatio(bool keep_aspect)
{
	LOG("SNES9xGraphics::SetAspectRatio(bool keep_aspect)\n");
	Settings.PS3KeepAspect = keep_aspect;
	SetViewports();
}

void SNES9xGraphics::Draw(int width, int height, uint16_t* screen)
{
	Clear();
	
	glBufferSubData(GL_TEXTURE_REFERENCE_BUFFER_SCE, 0, height * SNES9X_SCREEN_PITCH, screen);
	glTextureReferenceSCE(GL_TEXTURE_2D, 1, width, height, 0, GL_RGB5_A1, SCREEN_RENDER_TEXTURE_PITCH, 0);
	UpdateDimensions(width, height, width, height);
	
	glDrawArrays(GL_QUADS, 0, 4); 
	glFlush();
}

int32_t SNES9xGraphics::ChangeResolution(uint32_t resId, uint16_t refreshrateId)
{
	LOG("SNES9xGraphics::ChangeResolution(uint32_t resId, uint16_t refreshrateId)\n");
	int32_t ret;
	CellVideoOutState video_state;
	CellVideoOutConfiguration video_config;
	CellVideoOutResolution resolution;
	
	cellVideoOutGetState(CELL_VIDEO_OUT_PRIMARY, 0, &video_state);
	video_state.displayMode.refreshRates = refreshrateId;
	video_state.displayMode.resolutionId = resId;
	cellVideoOutGetResolution(video_state.displayMode.resolutionId, &resolution);
	
	memset(&video_config, 0, sizeof(CellVideoOutConfiguration));
	
	video_config.resolutionId = resId;
	video_config.format = CELL_VIDEO_OUT_BUFFER_COLOR_FORMAT_X8R8G8B8;
	video_config.aspect = video_state.displayMode.aspect;
	video_config.pitch = resolution.width << 2;
	
	Deinit();
	
	ret = cellVideoOutConfigure(CELL_VIDEO_OUT_PRIMARY, &video_config, NULL, 0);
	if(ret != CELL_OK)
	{
		return ret;
	}

	InitSNES9xGraphics(resId);
	PSGLGraphics::InitDbgFont();
	cellVideoOutGetState(CELL_VIDEO_OUT_PRIMARY, 0, &stored_video_state);
}

void SNES9xGraphics::UpdateDimensions(unsigned width, unsigned height, unsigned tex_width, unsigned tex_height)
{
	LOG("SNES9xGraphics::UpdateDimensions()\n");
	cgGLSetParameter2f(cg_video_size, width, height);
	cgGLSetParameter2f(cg_texture_size, tex_width, tex_height);
	cgGLSetParameter2f(cg_output_size, cg_viewport_width, cg_viewport_height);
}

void SNES9xGraphics::InitScreenQuad(int width, int height)
{
	LOG("SNES9xGraphics::InitScreenQuad(int width, int height)\n");
	screenQuad.v1.x = 0;
	screenQuad.v1.y = 0;
	screenQuad.v1.z = 0;

	screenQuad.v2.x = 0;
	screenQuad.v2.y = 1;
	screenQuad.v2.z = 0;

	screenQuad.v3.x = 1;
	screenQuad.v3.y = 1;
	screenQuad.v3.z = 0;

	screenQuad.v4.x = 1;
	screenQuad.v4.y = 0;
	screenQuad.v4.z = 0;

	screenQuad.t1.u = 0;
	screenQuad.t1.v = 1;

	screenQuad.t2.u = 0;
	screenQuad.t2.v = 0;

	screenQuad.t3.u = 1;
	screenQuad.t3.v = 0;

	screenQuad.t4.u = 1;
	screenQuad.t4.v = 1;

	memcpy(vertex_buf, ((float*)&screenQuad), 12 * sizeof(GLfloat));
	memcpy(vertex_buf + 128, ((float*)&screenQuad) + 12, 8 * sizeof(GLfloat));
}

void SNES9xGraphics::SetTextureFilter(bool keep_smooth)
{
	LOG("SNES9xGraphics::SetTextureFilter(bool keep_smooth)\n");
	Settings.PS3Smooth = keep_smooth;
	if (Settings.PS3Smooth)
	{
	   tex_filter = GL_LINEAR;
	}
	else
	{
		tex_filter = GL_NEAREST;
	}
	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, tex_filter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, tex_filter);
}

CGerror SNES9xGraphics::CheckCgError(int line)
{
	LOG("SNES9xGraphics::CheckCgError()\n");
	CGerror err = cgGetError();

	if (err != CG_NO_ERROR)
	{
		//FIXME: Add logging stuff
		LOG ("CG error:%d at line %d %s\n", err, line, cgGetErrorString (err));
	}

	return err;
}

CGprogram SNES9xGraphics::LoadShaderFromFile(CGprofile target, const char* filename, const char *entry)
{
	LOG("SNES9xGraphics::LoadShaderFromFile()\n");
	CGprogram id = cgCreateProgramFromFile(CgContext, CG_BINARY, filename, target, entry, NULL);
	if(!id)
	{
		//FIXME :Add logging stuff
		LOG("Failed to load shader program >>%s<<\nExiting\n", filename);
		CheckCgError(__LINE__);
	}

	return id;
}

CGprogram SNES9xGraphics::LoadShaderFromSource(CGprofile target, const char* filename, const char *entry)
{
	LOG("SNES9xGraphics::LoadShaderFromSource()\n");
	CGprogram id = cgCreateProgramFromFile(CgContext, CG_SOURCE, filename, target, entry, NULL);
	if(!id)
	{
		//FIXME :Add logging stuff
		LOG("Failed to load shader program >>%s<<\nExiting\n", filename);
		CheckCgError(__LINE__);
	}
	
	return id;
}

int32_t SNES9xGraphics::InitShader(const char* filename)
{
   LOG("SNES9xGraphics::InitShader()\n");
   CGprogram id = LoadShaderFromSource(CG_PROFILE_SCE_FP_RSX, filename, "main_fragment");
   if (id)
   {
      FragmentProgram = id;
      cgGLBindProgram(FragmentProgram);
      ModelViewProj_cgParam = cgGetNamedParameter(VertexProgram, "modelViewProj");
      cgGLSetStateMatrixParameter(ModelViewProj_cgParam, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
      cg_video_size = cgGetNamedParameter(FragmentProgram, "IN.video_size");
      cg_texture_size = cgGetNamedParameter(FragmentProgram, "IN.texture_size");
      cg_output_size = cgGetNamedParameter(FragmentProgram, "IN.output_size");
      return CELL_OK;
   }
   return !CELL_OK;
}

void SNES9xGraphics::InitCG()
{
   LOG("SNES9xGraphics::InitCG()\n");
   CgContext = cgCreateContext();
   cgRTCgcInit();

   const char *shader = (strlen(Settings.PS3CurrentShader) > 0) ? Settings.PS3CurrentShader : DEFAULT_PROGRAM_SOURCE;
   VertexProgram = LoadShaderFromSource(CG_PROFILE_SCE_VP_RSX, shader, "main_vertex");
   FragmentProgram = LoadShaderFromSource(CG_PROFILE_SCE_FP_RSX, shader, "main_fragment");

   cgGLEnableProfile(CG_PROFILE_SCE_VP_RSX);
   cgGLEnableProfile(CG_PROFILE_SCE_FP_RSX);
   cgGLBindProgram(VertexProgram);
   cgGLBindProgram(FragmentProgram);

   ModelViewProj_cgParam = cgGetNamedParameter(VertexProgram, "modelViewProj");
   cg_video_size = cgGetNamedParameter(FragmentProgram, "IN.video_size");
   cg_texture_size = cgGetNamedParameter(FragmentProgram, "IN.texture_size");
   cg_output_size = cgGetNamedParameter(FragmentProgram, "IN.output_size");
}

int32_t SNES9xGraphics::PSGLInit()
{
	LOG("SNES9xGraphics::PSGLInit()\n");
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_SCISSOR_TEST);
	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	glDisable(GL_FOG);
	glDisable(GL_DITHER);
	glShadeModel(GL_FLAT);
	glEnable(GL_VSYNC_SCE);
	glEnable(GL_TEXTURE_2D);
	
	InitCG();
	SetViewports();

	gl_buffer = (uint8_t*)memalign(128, SCREEN_RENDER_TEXTURE_HEIGHT * SCREEN_RENDER_TEXTURE_PITCH); // Allocate memory for texture.
   memset(gl_buffer, 0, SCREEN_RENDER_TEXTURE_HEIGHT * SCREEN_RENDER_TEXTURE_PITCH);
	vertex_buf = (uint8_t*)memalign(128, 256);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glGenBuffers(2, vbo);

	glBindBuffer(GL_TEXTURE_REFERENCE_BUFFER_SCE, vbo[0]);
	glBufferData(GL_TEXTURE_REFERENCE_BUFFER_SCE, SCREEN_RENDER_TEXTURE_HEIGHT * SCREEN_RENDER_TEXTURE_PITCH, gl_buffer, GL_STREAM_DRAW);

	glTextureReferenceSCE(GL_TEXTURE_2D, 1, SCREEN_RENDER_TEXTURE_WIDTH, SCREEN_RENDER_TEXTURE_HEIGHT, 0, GL_RGB5_A1, SCREEN_RENDER_TEXTURE_PITCH, 0);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

	SetTextureFilter(Settings.PS3Smooth);

	// PSGL doesn't clear the screen on startup, so let's do that here.
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glColor4f(1.0, 1.0, 1.0, 1.0);
	Clear();
	psglSwap();

   // Use some initial values for the screen quad.
	InitScreenQuad(SCREEN_RENDER_TEXTURE_WIDTH, SCREEN_RENDER_TEXTURE_HEIGHT);

	glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
	glBufferData(GL_ARRAY_BUFFER, 256, vertex_buf, GL_STATIC_DRAW);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, 0);
	glTexCoordPointer(2, GL_FLOAT, 0, (void*)128);

	return CELL_OK;
}

int SNES9xGraphics::InitSNES9xGraphics()
{
	LOG("SNES9xGraphics::InitSNES9xGraphics()\n");
	int32_t ret;
	PSGLGraphics::Init(NULL, Settings.PS3PALTemporalMode60Hz);
	ret = PSGLInit();

	if (ret == CELL_OK)
	{
		cellVideoOutGetState(CELL_VIDEO_OUT_PRIMARY, 0, &stored_video_state);
	}

	return ret;
}

int SNES9xGraphics::InitSNES9xGraphics(uint32_t resId)
{
	LOG("SNES9xGraphics::InitSNES9xGraphics(uint32_t resId)\n");
	int32_t ret;
	PSGLGraphics::Init(resId, Settings.PS3PALTemporalMode60Hz);
	ret = PSGLInit();

	/*
	if (ret == CELL_OK)
	{
		cellVideoOutGetState(CELL_VIDEO_OUT_PRIMARY, 0, &stored_video_state);
	}
	*/

	return ret;
}
