#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timer.h>
#include <sys/return_code.h>
#include <sys/process.h>
#include <cell/sysmodule.h>
#include <cell/cell_fs.h>
#include <cell/pad.h>
#include <cell/dbgfont.h>
#include <cell/audio.h>
#include <stddef.h>
#include <math.h>
#include <sysutil/sysutil_sysparam.h>
#include <sys/spu_initialize.h>

#include "emulator.h"
#include "menu.h"

#include "snes9x.h"
#include "memmap.h"
#include "apu/apu.h"
#include "gfx.h"
#include "snapshot.h"
#include "controls.h"
#include "cheats.h"
#include "movie.h"
#include "logger.h"
#include "display.h"
#include "conffile.h"

#include "cellframework/logger/Logger.h"
#include "cellframework/input/cellInput.h"
#include "SNES9xGraphics.h"
//#include "emulator_graphics.h"
//#include "emulator_audio.h"
#include "cellframework/audio/rsound.hpp"
#include "emulator_implementation.h"
#include "cellframework/utility/OSKUtil.h"

CellInputFacade* CellInput;
SNES9xGraphics* Graphics;
SYS_PROCESS_PARAM(1001, 0x10000);

// mode the main loop is in
Emulator_Modes mode_switch = MODE_MENU;

// is a ROM running
bool is_running;

// is emulator loaded?
bool emulator_loaded = false;

// needs settings loaded
bool load_settings = true;

// need to load the current rom
bool need_load_rom = false;

// current rom being emulated
char* current_rom = NULL;

Audio::Stream<int16_t> *CellAudio;
OSKUtil *oskutil;

// controller mode to run emulator in
static Emulator_ControllerModes controller_settings = NO_JOYSTICKS;


// debug text for output during emulation loop
const char *debug_text;

// current save state id
unsigned int current_save_state = 0;
unsigned int current_cheat_position = 0;

unsigned int Emulator_CurrentSaveStateSlot()
{
	return current_save_state;
}

void Emulator_DecrementCurrentSaveStateSlot()
{
	if(Emulator_CurrentSaveStateSlot() > 0)
	{
		current_save_state--;
	}
}

void Emulator_IncrementCurrentSaveStateSlot()
{
	if(Emulator_CurrentSaveStateSlot() < 9)
	{
		current_save_state++;
	}
}

unsigned int Emulator_CurrentCheatPosition()
{
	return current_cheat_position;
}

void Emulator_DecrementCurrentCheatPosition()
{
	if(Emulator_CurrentCheatPosition() > 0)
	{
		current_cheat_position--;
	}
}

void Emulator_IncrementCurrentCheatPosition()
{
		current_cheat_position++;
}

void Emulator_SetCurrentSaveStateSlot(int set_current_save_state_slot)
{
	if(set_current_save_state_slot <= 9)
	{
		current_save_state = set_current_save_state_slot;
	}
}

int Emulator_LoadSettings()
{
	return 1;
}

int Emulator_SaveSettings()
{
	if(Emulator_Implementation_SaveSettings())
	{
		return 1;
	}
	else
	{
		return 0;
	}

}


void Emulator_SetControllerMode(Emulator_ControllerModes m)
{
	controller_settings = m;
}


Emulator_ControllerModes Emulator_GetControllerMode()
{
	return controller_settings;
}


bool Emulator_IsInitialized()
{
	return emulator_loaded;
}


bool Emulator_IsROMLoaded()
{
	return current_rom != NULL && need_load_rom == false;
}


bool Emulator_ROMRunning()
{
	return is_running;
}

char * Emulator_GetROM()
{
	return current_rom;
}

void Emulator_SetDebugMessage(const char * value)
{
	debug_text = value;
}


void Emulator_SwitchMode(Emulator_Modes m)
{
	mode_switch = m;
}


void Emulator_OSKStart(const wchar_t* msg, const wchar_t* init)
{
	oskutil->Start(msg,init);
}

const char * Emulator_OSKOutputString()
{
	return oskutil->OutputString();
}

void Emulator_Shutdown()
{
	Emulator_SaveSettings();
LOG("delete Graphics\n");
   delete Graphics;
LOG("delete CellAudio\n");
   delete CellAudio;
LOG("delete oskutil\n");
   delete oskutil;

	cellSysmoduleUnloadModule(CELL_SYSMODULE_FS);
	cellSysmoduleUnloadModule(CELL_SYSMODULE_IO);
	cellSysutilUnregisterCallback(0);


	// do any clean up... save stuff etc
	// ...

	// force exit
	exit(0);
}


void Emulator_RequestLoadROM(char* rom, bool forceReload)
{
	if (forceReload || current_rom == NULL || strcmp(rom, current_rom) != 0)
	{
		if (current_rom != NULL)
		{
			free(current_rom);
		}

		current_rom = strdup(rom);
		need_load_rom = true;
	}
	else
	{
		need_load_rom = false;
	}
}

void Emulator_SaveSRAM()
{
	Emulator_Implementation_SaveSRAM();
}


void Emulator_StopROMRunning()
{
	is_running = false;
}

void Emulator_StartROMRunning()
{
	Emulator_SwitchMode(MODE_EMULATION);
}

/*
 * Callback for PS3 System operations
 */
void sysutil_exit_callback (uint64_t status, uint64_t param, void *userdata)
{
	(void) param;
	(void) userdata;

	switch (status)
	{
		case CELL_SYSUTIL_REQUEST_EXITGAME:
			MenuStop();
			Emulator_StopROMRunning();
			Emulator_SwitchMode(MODE_EXIT);
			break;
		case CELL_SYSUTIL_DRAWING_BEGIN:
		case CELL_SYSUTIL_DRAWING_END:
			break;
		case CELL_SYSUTIL_OSKDIALOG_LOADED:
			break;
		case CELL_SYSUTIL_OSKDIALOG_FINISHED:
			oskutil->Stop();
			break;
		case CELL_SYSUTIL_OSKDIALOG_UNLOADED:
			oskutil->Close();
			break;
		default:
			break;
	}
}


bool Emulator_InitSettings ()
{
	Emulator_Implementation_InitSettings();
}


bool Emulator_Init()
{
	return Emulator_Implementation_InitSystem(controller_settings);
}

bool Emulator_Initialize()
{
	emulator_loaded = Emulator_Init();
}


void Emulator_Inputloop ()
{
	Emulator_Implementation_Inputloop();
}

#if 1
void Emulator_PlayAudio ()
{
   // Audioport requires 256 frames of audio at a time.
#define AUDIO_BLOCK_SAMPLES (256)
#define AUDIO_CHANNELS (2)
	int16_t samples[AUDIO_BLOCK_SAMPLES * AUDIO_CHANNELS];

	while (S9xGetSampleCount() >= AUDIO_BLOCK_SAMPLES * AUDIO_CHANNELS && CellAudio->write_avail() >= AUDIO_BLOCK_SAMPLES * AUDIO_CHANNELS)
	{
		S9xMixSamples((uint8 *)samples, AUDIO_BLOCK_SAMPLES * AUDIO_CHANNELS);
      CellAudio->write(samples, AUDIO_BLOCK_SAMPLES * AUDIO_CHANNELS);
	}
}
#endif

void Emulator_ToggleSound()
{
	if(CellAudio)
	{
		delete CellAudio;
	}
	if((Settings.RSoundEnabled) && (strlen(Settings.RSoundServerIPAddress) > 0))
	{
		CellAudio = new Audio::RSound<int16_t>(Settings.RSoundServerIPAddress, 2, AUDIO_INPUT_RATE);
      // If we couldn't connect, fall back to normal audio...
      if (!CellAudio->alive())
      {
         delete CellAudio;
         CellAudio = new Audio::AudioPort<int16_t>(2, AUDIO_INPUT_RATE);
         Settings.RSoundEnabled = false;
         Graphics->Clear();
         cellDbgFontPuts(0.2f, 0.4f, FONT_SIZE, 0xffffffff, "Couldn't connect to RSound server.\nFalling back to regular audio.");
         Graphics->RenderDbgFont();
         Graphics->Swap();
         sys_timer_usleep(3000000);
      }
	}
	else
	{
		CellAudio = new Audio::AudioPort<int16_t>(2, AUDIO_INPUT_RATE);
	}
}

void Emulator_Start()
{
	if(need_load_rom)
	{
		if (!Emulator_Initialize())	
		{
			Settings.Paused = TRUE;
		}
		else
		{
			need_load_rom = false;
		}
	}

	
	S9xSetSoundMute(FALSE);

	//bring down dbgfont
	Graphics->DeInitDbgFont();
	while (is_running)
	{
		Emulator_Inputloop();

		Emulator_PlayAudio();
		if (S9xSyncSound())
		{
			S9xMainLoop();
         Graphics->Swap();
		}

		cellSysutilCheckCallback();
	}

	Graphics->InitDbgFont();
	Emulator_SaveSRAM();
	S9xSetSoundMute(TRUE);
	is_running = true;
}

void Emulator_GraphicsInit(uint32_t resId)
{
	LOG("Emulator_GraphicsInit(resId)\n");
	if(Graphics == NULL)
	{
		Graphics = new SNES9xGraphics();
	}

	LOG("Emulator_GraphicsInit->InitSNES9xGraphics(resId)\n");
	Graphics->InitSNES9xGraphics(resId);

	LOG("Emulator_GraphicsInit->InitDbgFont()\n");
	Graphics->InitDbgFont();
}

int main (void) {
   sys_spu_initialize(6, 1);
	cellSysmoduleLoadModule(CELL_SYSMODULE_FS);
	cellSysmoduleLoadModule(CELL_SYSMODULE_IO);

	cellSysutilRegisterCallback(0, sysutil_exit_callback, NULL);
   LOG_INIT();
   LOG("LOG INIT\n");
	
	CellInput = new CellInputFacade();
	CellInput->Init();

	oskutil = new OSKUtil();

	// load SNES9x settings before calling rom_menu() and Graphics_Init() so they are
	// initialized correctly, in case settings screen is accessed before a ROM is loaded

	if (load_settings)
	{
		Emulator_InitSettings();
		load_settings =  false;
	}

	Graphics = new SNES9xGraphics();
	Graphics->InitSNES9xGraphics();
	Graphics->SetOverscan(Settings.PS3OverscanEnabled,(float)Settings.PS3OverscanAmount/100);
	Graphics->InitDbgFont();
	Emulator_ToggleSound();
   
	is_running = true;

	while(1)
	{
		switch(mode_switch)
		{
			case MODE_MENU:
				MainMenu();
				break;
			case MODE_EMULATION:
            CellAudio->unpause();
				Emulator_Start();
            CellAudio->pause();
				break;
			case MODE_EXIT:
				Emulator_Shutdown();
				return 0;
		}
	}

	return 0;
}

